import std.stdio ;
import std.string ;

enum UNDOSTATE{
  NONE,
  TOP,
  MEDIUM,
  BOTTOM
}

class UndoBuffer{
  Object[]  objbuf ;
  bool[]    affect ;
  int[]     difbuf ;
  bool      prev_affect ;
  bool      changedData ;
  int       topptr ;
  int       curptr ;
  int       saveptr ;
  const int GCminimize_threshold=30 ;
  int       GCminimize=1 ;

  this(int level){
    objbuf.length=level ;
    affect.length=level ;
    difbuf.length=level ;
    clear() ;
  }
  
  ~this(){
    delete objbuf ;
    delete affect ;
    delete difbuf ;
  }

  int setSavePoint(){
    saveptr=curptr ;
    prev_affect=false ;
    return(0) ;
  }
  
  int push(Object obj,bool affectflag){
    GCminimize = (GCminimize+1) % GCminimize_threshold ;
    if( GCminimize==0 ) core.memory.GC.minimize() ;

    if( curptr==objbuf.length-1 ){
      if( saveptr==0 ){
	if( difbuf[saveptr]!=difbuf[curptr] ) prev_affect=true ;
      }else{
	if( affect[0]==true ) prev_affect=true ;
      }

      delete objbuf[0] ;
      for( int i=0 ; i<objbuf.length-1 ; i++ ){
	objbuf[i]=objbuf[i+1] ;
	affect[i]=affect[i+1] ;
      }
      curptr=cast(int)(objbuf.length)-2 ;
      if( saveptr>=0 ) saveptr-- ;
    }

    curptr++ ;
    topptr=curptr ;
    objbuf[topptr]=obj ;
    affect[topptr]=affectflag ;

    int diffcount=0 ;
    for( int i=0 ; i<=topptr ; i++ ){
      if( affect[i]==true ) diffcount++ ;
      difbuf[i]=diffcount ;
    }
    return(0) ;
  }

  int clear(){
    objbuf[]=null ;
    affect[]=false ;
    prev_affect =false ;
    topptr= -1 ;
    curptr= -1 ;
    saveptr=-1 ;
    return(0) ;
  }
  
  Object undo(){
    if( curptr==0 ) return(null) ;
    curptr-- ;
    return(objbuf[curptr]) ;
  }
  
  Object redo(){
    if( topptr==curptr ) return(null) ;
    curptr++ ;
    return(objbuf[curptr]) ;
  }

  UNDOSTATE getUndoState(){
    if( curptr==-1 && topptr==-1 ) return(UNDOSTATE.NONE) ;
    if( curptr==0  && topptr==0  ) return(UNDOSTATE.NONE) ;
    if( curptr==topptr           ) return(UNDOSTATE.TOP) ;
    if( curptr==0                ) return(UNDOSTATE.BOTTOM) ;
    return(UNDOSTATE.MEDIUM) ;
  }

  bool hasChangedData(){
    if( saveptr>=0 ){
      if( difbuf[saveptr]!=difbuf[curptr] ) return(true) ;
      return(false) ;
    }else{
      if( prev_affect==true ) return(true) ;
      if( (curptr>=0 && difbuf[0]==0 && difbuf[0]==difbuf[curptr]) || curptr<0 ) return(false) ;
      return(true) ;
    }
  }

  string dgbprint(){
    string affectmark="" ;
    for( int i=0 ; i<affect.length ; i++ ){
      if( affect[i]==true ) affectmark = affectmark ~ "X" ;
      else affectmark = affectmark ~ "-" ;
    }
    return(format("saveptr=%d curptr=%d topptr=%d prev_affect=%s %s res=%s",
		  saveptr,curptr,topptr,prev_affect,affectmark, hasChangedData())) ;
  }
  
}

/********************************************************************************
メモ

  ■使用方法

    (1)変更点で push()で差分データと 更新の影響有無を指定する。
    (2-a)undo()で最後にpush()で登録した一つ前以前の差分データを取り出す。
    (2-b)redo()でundo()で変更したポインタを戻す。
    (3)(1)に戻る。
    (4)セーブポイントでsetSavePoint()を実行する。

    push()でのaffectflagは、例えば「データの選択」の様にデータ自体が変更
    される訳では無いが操作を記録したい場合にはfalseを指定する。

    push()時のオブジェクトについては規定しない。元状態の完全なデータのコピー
    かも知れないし、差分データかも知れないし、差分操作かも知れないが、
    Undo/Redo元に戻すのに必要なデータ/操作を記録したオブジェクト(のインス
    タンス)を指定する。

  ■UndoBuffer構造について

    UndoBufferはスタックのような形状で保持する。10段のバッファサイズを保持
    した場合のイメージを示す。

     ポインタ   -1   0   1   2   3   4   5   6   7   8   9
	      +---┌─┬─┬─┬─┬─┬─┬─┬─┬─┬─┐
          obj |   │  │  │  │  │  │  │  │  │  │  │
              +---├─┼─┼─┼─┼─┼─┼─┼─┼─┼─┤
       affect |   │  │  │  │  │  │  │  │  │  │  │
              +---└─┴─┴─┴─┴─┴─┴─┴─┴─┴─┘

    バッファはコンストラクタで確保したサイズを途中で変更する事はせず、
    ポインタのみで 書き込み/読み出し/最大位置 を管理する。
    サイズを越える場合は [1 .. 9]の領域を[0 .. 8]にコピーした後、[9]に
    push()される。

  ■セーブポイントからUndo/Redo時差分の有無を検出するアルゴリズムについて

    次のようなバッファ状態を例にセーブポイントからの差分の有無を検査する方法を
    示す。

    1,4,8でデータに変化のある更新を行い、セーブポイントを4に設定したとする。
    Undoの現在位置はCP(curptr)はひとまず7とする。

                -1   0   1   2   3   4   5   6   7   8   9
	      +---┌─┬─┬─┬─┬─┬─┬─┬─┬─┬─┐   －:false
       affect |   │－│×│－│－│×│－│－│－│×│－│   ×:true
              +---└─┴─┴─┴─┴─┴─┴─┴─┴─┴─┘
                                    SV          CP

    このaffectバッファより、差分を検査するためのdifbuf(差分バッファ)を生成する。
    affectバッファを順に走査し、affect[x]==trueのときに値をインクリメントした
    内容をdifbufには保持する。difbufはpush()の度に再構築される。

                -1   0   1   2   3   4   5   6   7   8   9
	      +---┌─┬─┬─┬─┬─┬─┬─┬─┬─┬─┐   －:false
       affect |   │－│×│－│－│×│－│－│－│×│－│   ×:true
              +---└─┴─┴─┴─┴─┴─┴─┴─┴─┴─┘
	          ┌─┬─┬─┬─┬─┬─┬─┬─┬─┬─┐
       difbuf     │ 0│ 1│ 1│ 1│ 2│ 2│ 2│ 2│ 3│ 3│
                  └─┴─┴─┴─┴─┴─┴─┴─┴─┴─┘
                                    SV          CP

    セーブポイントの difbuf[SV] の値と 現在位置のdifbuf[CP]の値が同じならば、
    セーブポイントから更新が無く、違っていれば更新があると判定できる。

    例1 : CPが7の時、セーブポイント4からデータに差の出る操作は5,6,7で行われて
          いないので、データに変更無しという事が判る。
    例2 : CPが9だった時、セーブポイントから8の位置でデータに変更がある事が判る。
    例3 : CPが2だった時、1の更新の後で4の更新をしてセーブているので、3以前は
          4と異なったデータだと言える。4でセーブしたデータを1のデータに戻す
          事になるので、Undoするとデータが変更されると判る。

********************************************************************************/
